/******************************************************************************************/
/* Coypright 2013, Winfried@sunkiddance.de aka  =-WDI-=, 21.12.2013                       */
/*                                                                                        */
/* This program is free software: you can redistribute it and/or modify                   */
/*     it under the terms of the GNU General Public License version 3 as published by     */
/*     the Free Software Foundation.                                                      */
/*                                                                                        */
/*     This program is distributed in the hope that it will be useful,                    */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of                     */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                      */
/*     GNU General Public License for more details.                                       */
/*                                                                                        */
/*     You should have received a copy of the GNU General Public License                  */
/*     along with this program.  If not, see <http://www.gnu.org/licenses/>.              */
/******************************************************************************************/
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
 
#define bool int
#define true 1
#define false 0

#define SECTORSIZE 2048 /* DVD-Video sector size is 2048 bytes */

/*-------------------------------------------------------------------------------------------*/
/* Determine size of file                                                       	     */
/* const char* filename: filename                                               	     */
/* Code snippet shamelessly stolen from                                         	     */           
/* http://stackoverflow.com/questions/8236/how-do-you-determine-the-size-of-a-file-in-c?lq=1 */
/*-------------------------------------------------------------------------------------------*/
off_t fsize(const char *filename) 
{
    struct stat st;

    if (stat(filename, &st) == 0)
        return st.st_size;

    fprintf(stderr, "error: Cannot determine size of %s: %s\n",
            filename, strerror(errno));

    return -1;
} /* end of fsize() */

/*------------------------------------------------------------------------------*/
/* Check whether a file exists                                                  */
/* const char filename: file name of file to check                              */
/* Code snippet shamelessly stolen from                                         */
/* http://stackoverflow.com/questions/230062/whats-the-best-way-to-check-if-a-file-exists-in-c-cross-platform */
/*------------------------------------------------------------------------------*/ 
int file_exist (const char *filename)
{
  struct stat buffer;   
  return (stat (filename, &buffer) == 0);
}

/*------------------------------------------------------------------------------*/
/* Remove file if if exists                                                     */
/* const* char fileName: a file name                                            */
/*------------------------------------------------------------------------------*/
void removeFile(const char* fileName)
{
  if(file_exist(fileName))
    {
      if(remove(fileName) != 0) 
	{
	  fprintf(stderr, "error: can't remove output VOB file' %s': %s\n", 
		  fileName, strerror(errno));
	}
    }
}

/*------------------------------------------------------------------------------*/
/* Print the usage message                                                      */
/*------------------------------------------------------------------------------*/
void usage(void)
{
      fprintf(stderr, "VobCut 0.1 - VOB Cutting Utility\n");
      fprintf(stderr, "Usage: vobcut inVOB outVOB startsec endsec\n");
}

/*------------------------------------------------------------------------------*/
/* Issue error message via stderr                                               */
/* const char* message: format string for the message to print to stderr        */
/* const char* arg1: first % argument for the format string                     */
/* const char* arg2: second % argument for the format string                    */
/* bool showUsage: show usage information if true                               */
/*------------------------------------------------------------------------------*/
void handleError(const char* message, const char* arg1, const char* arg2, bool showUsage)
{
  if(message == 0)
    {
      fprintf(stderr, "handleError: Please provide at least an error message text!\n");
      return;
    }
  if(arg1 == 0 && arg2 == 0)
    {
      fputs(message, stderr);
    }
  else if((arg1 != 0 && arg2 == 0) || (arg1 == 0 && arg2 != 0))
    {
      fprintf(stderr, message, arg1 == 0 ? arg2 : arg1);
    }
  else
    {
      fprintf(stderr, message, arg1, arg2);
    }
  if(showUsage) usage();
}

/*------------------------------------------------------------------------------*/
/* Cut a part of an DVD-Video VOB file and write it to a new file.              */
/* const char* inputVOBFile: a file name for a valid DVD-Video VOB file         */
/* const char* outputVOBFile: a file name for a valid DVD-Video VOB file        */
/* const int startSec: DVD-Video sector number (0...524287)                     */
/* const int endSec: DVD-Video sector number (0...524287)                       */
/* Note that a valid VOB file is at most 1 GB large. The cutting tool doesn't   */
/* enforce this constraint however.                                             */
/*------------------------------------------------------------------------------*/
int _SplitVOB(const char* inputVOBFile, const char* outputVOBFile, const int startSec, const int endSec)
{
  FILE *inputHandle, *outputHandle;
  int result, numSec, block;
  off_t sizeInputVOB;
  unsigned char	buffer[SECTORSIZE];
  char convBuffer[64];

  result = 1;
  if((inputHandle = fopen(inputVOBFile, "rb")) == 0) 
    {
      handleError("error: can't open input file '%s: %s'\n", inputVOBFile, strerror(errno), false);
      return result;
    }

  /* Determine file size of input VOB */
  if((sizeInputVOB=fsize(inputVOBFile)) == -1)
    {
      handleError("error: can't determine size of file '%s': %s!\n", inputVOBFile, strerror(errno), false);
      fclose(inputHandle);
      return result;
    }

  /* Check whether the input VOB file size is a multiple of sector size */
  if(sizeInputVOB % SECTORSIZE != 0)
    {
      snprintf(convBuffer, sizeof(convBuffer), "%d", SECTORSIZE);
      handleError("Size of input VOB '%s': must be a multiple of %s!\n", inputVOBFile, convBuffer, false); 
      fclose(inputHandle);
      return result;
    }

  /* Check whether the end sector is within the file size */
  if((numSec = sizeInputVOB / SECTORSIZE - 1) < endSec)
    {
      char convBuffer1[64];

      snprintf(convBuffer, sizeof(convBuffer), "%d", endSec);
      snprintf(convBuffer1, sizeof(convBuffer1), "%d", numSec);
      handleError("End sector '%s' is out of bound, can be at most '%s'!\n", convBuffer, convBuffer1, false);
      fclose(inputHandle);
      return result;
    }      

  if(fseek(inputHandle, (long) SECTORSIZE*startSec, SEEK_SET) != 0 )
    {
      handleError("error: can't jump to start sector in file '%s': %s\n", inputVOBFile, strerror(errno), false);
      fclose(inputHandle);
      return result;
    }

  if((outputHandle = fopen(outputVOBFile, "wb")) == 0)
    {
      handleError("error: can't open output file '%s': %s\n", outputVOBFile, strerror(errno), false);
      fclose(inputHandle);
      return result;
    }

  for(block = startSec; block <= endSec; block++)
    {
      if(fread(buffer, 1, sizeof buffer, inputHandle) != sizeof buffer)
	{
	  handleError("error: can't read enough date in file '%s': %s\n", inputVOBFile, strerror(errno), false);
	  fclose(inputHandle);
	  fclose(outputHandle);
	  return result;
	}
      else if(fwrite(buffer, 1, sizeof buffer, outputHandle) != sizeof buffer)
	{
	  handleError("error: can't write enough data to output file' %s': %s\n", outputVOBFile, strerror(errno), false);
	  fclose(inputHandle);
	  fclose(outputHandle);
	  removeFile(outputVOBFile);
	  return result;
	}
    } /* end for */

  fclose(inputHandle);
  fclose(outputHandle);
  result = 0;
  puts("Cutting, done.");
  return result;
} /* end _SplitVOB() */

/*------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
  char sourceFile[255], targetFile[255]; /* the length chosen is arbitrary */
  int startSec, endSec;
  size_t lenArg;
  
  if(argc != 5) 
    {
      handleError("Wrong number of arguments!\n", "", "", true);
      return 1;
    }
  if(sscanf(argv[1], "%254s", sourceFile) != 1)
    {
      handleError("Error parsing sourcefileVOB '%s' (string expected)\n", argv[1], "", true);
      return 1;
    }
  if(sscanf(argv[2], "%254s", targetFile) != 1)
    {
      handleError("Error parsing targetfileVOB) '%s' (string expected\n)", argv[2], "", true);
      return 1;
    }
  lenArg=strlen(argv[3]);
  if(lenArg > 10)
    {
      handleError("Error parsing startsector (argument too long, must be < 11 chars)\n", "", "", true);
      return 1;
    }
  if(sscanf(argv[3], "%10d", &startSec) != 1)
    {
      fprintf(stderr, "Error parsing startsector '%s'(positive interger expected)\n", argv[3]);
      usage();
      return 1;
    }
  lenArg=strlen(argv[4]);
  if(lenArg > 10)
    {
      handleError("Error parsing endsector (argument too long, must be < 11 chars)\n", "", "", true);
      return 1;
    }
  if(sscanf(argv[4], "%10d", &endSec) != 1)
    {
      handleError("Error parsing endsector '%s' (positive interger expected)\n", argv[4], "", true);
      return 1;
    }
  if(startSec < 0 || endSec < 0 || startSec > endSec)
    {
      handleError("Error: 'endsector' can't be smaller than 'startsector' and sector number has to be at least 0!\n", "", "", true);
      return 1;
    }

  return _SplitVOB(sourceFile, targetFile, startSec, endSec);

} /* end of main() */ 
